//////////////////////////////////////////////////////////////////////////////
// LINK: Generic Graph Tool and Class Library                               
//      Function name: ShortestPath.cc                                      
//                                                                            
//      Synopsis:  Computes the single source shortest path of a graph.
//                                                                            
//      Description:             					      
//                                                                            
//      Creation:  Patricia Fasel pkf@c3.lanl.gov
//                                                                            
//      Routines used: BinaryHeaps
//                                                                            
//      Related files: Graph.h, Vertex.h, Edge.h, Attribute.h, BinaryHeap.h
//
//      Test suite:                                                           
//                                                                            
//      User documentation:                                                   
//                                                                            
//      Development History:                                                  
//           4/19/93 - murphy@c3.lanl.gov incoporated Binary Heaps to
//                     improve asymptotic running time
//
//           6/3/93  - murphy@c3.lanl.gov fixed it so it would work with
//                     directed graphs 
//
//           7/18/93 - ejohnson@c3.lanl.gov fixed problem that occurs when
//                     opposing edges exist in binary graph
//
//           7/21/93 - ejohnson@c3.lanl.gov added check when adding distance
//                     attribute to ensure that it doesn't already exist
//
//           8/16/93 - ejohnson@c3.lanl.gov changed initialization so
//                     that it works on directed and undirected graphs
//
//      Testing History:                                                      
//           date  - full@email.address.xxx  description of tests done
//                                                                           
//      Code Review:                                                          
//           date  - full@email.address.xxx  any notes you may have 
//
//      Bugs and Deficiencies:                                               
//                                                                            
//////////////////////////////////////////////////////////////////////////////


#include <stdio.h>
#include <LINK/graph/Graph.h>
#include <LINK/graph/Vertex.h>
#include <LINK/graph/Edge.h>
#include <LINK/graph/Attribute.h>
#include <LINK/basic/BinaryHeap.h>


//
// for the source vertex, record all distances from it, using edge weights
//
void
initSingleInput(Graph* graph, Vertex* source, BinaryHeap<int,Vertex*> &heap)
{
    ContainerNode* heapPtr;
    int large_int = LARGE_INT;

    Vertex** vertices = graph->vertexStart();
    int order = graph->order();
    for (int i=0; i < order; i++) {

	// source vertex has a distance of 0
	if (vertices[i] == source) {
            heapPtr = heap.insert(0, vertices[i]);

	// adjacent vertices have a distance of the edge weight
        } else {
	    Set<Edge*> in_edges = graph->inIncidentEdges(vertices[i]);
	    Iterator<Edge*> get_edge(&in_edges);
	    Edge* edge;
 	    Bool found = FALSE;
	    while (get_edge(edge) && !found) {
	        if (edge->hasVertex(source)) {
		    int weight;
		    getAttribute((GraphObject*) edge, "weight", weight);
                    heapPtr = heap.insert(weight, vertices[i]);
	            found = TRUE;
	        }
	    }
	    if (!found)
                heapPtr = heap.insert(large_int, vertices[i]);
	}
	setAttribute((GraphObject*) vertices[i], "heap_info", heapPtr);
    }
}


//
// Dijkstra's single source shortest path algorithm 
// modified to operate on hypergraphs
//
void
ShortestPath(Graph* graph, Vertex* source_vertex)
{
    static char SccsId[] = "@(#)ShortestPath.cc	3.4\t9/11/95";
    printf("%s\n",SccsId);

    Vertex* vertex;
    ContainerNode* heapPtr;
    Attribute<ContainerNode*> bin_heap(graph, "heap_info", (ContainerNode*) 0);

    // create the binary heap to store the distance data
    int num_vertices = graph->order();    
    BinaryHeap<int,Vertex*> heap(num_vertices);

    // initialize the distances of all vertices from the source vertex
    initSingleInput(graph, source_vertex, heap);

    // remove the source vertex from the processing list
    heap.extractMin();   

    // iterate through the processing list until empty, selecting the min vertex
    for (int j=1; j< num_vertices; j++) {
	BinaryHeapNode<int,Vertex*>* heap_node = heap.extractMin();
        vertex = heap_node->item;
        int u_dist = heap_node->key;

	// calculate new distances for all neighbors of the current min vertex
	Set<Edge*> out_edges = graph->outIncidentEdges(vertex);
	Iterator<Edge*> get_incident_edge(&out_edges);
	Edge* inc_edge;
	while (get_incident_edge(inc_edge)) {

	    // every outIncidentEdge leads to neighbor vertices of vertex
	    Iterator<Vertex*> get_neighbor(inc_edge->vertices());
	    Vertex* neighbor_vertex;
	    while (get_neighbor(neighbor_vertex))

		// change distances for all neighbors of vertex
		if (vertex != neighbor_vertex) {

		    // get current distance of the neighbor
		    getAttribute((GraphObject*) neighbor_vertex, 
			"heap_info", heapPtr);
		    int v_dist = ((BinaryHeapNode<int,Vertex*>*) heapPtr)->key;

		    // alter that distance according to incident edge weight
		    int e_weight;
		    getAttribute((GraphObject*) inc_edge, "weight", e_weight);
		    if (v_dist > u_dist + e_weight)
			heap.decreaseKey(heapPtr, u_dist+e_weight);
		}
	}
    }

    // put the result into "distance" attribute belonging to each vertex
    Vertex** vertices = graph->vertexStart();
    int order = graph->order();
    for (int i=0; i < order; i++) {
	getAttribute((GraphObject*) vertices[i], "heap_info", heapPtr);
        int dist = ((BinaryHeapNode<int,Vertex*>*) heapPtr)->key;
	setAttribute((GraphObject*) vertices[i], "distance", dist);
    }
}
